#ifndef CONSTRAINTDB_HPP_
#define CONSTRAINTDB_HPP_

#include    "Types.hpp"
#include    "Statistics.hpp"
#include    "Settings.hpp"
#include    <vector>
#include	<list>


using std::vector;

class ConstraintDB;


//const uint32_t INVALID_LEVEL     = ~zero32;

const uint8_t VAR_VALUE_MASK    	= 3;                // = 00000011
const uint8_t VAR_VALUE_MASK_INV 	= ~VAR_VALUE_MASK; //  = 11111100
const uint8_t VAR_ALL_VAL_MASK  	= 15;   	        // = 00001111
const uint8_t VAR_HOOK_MASK     	= 3  << 6;          // = 11000000
const uint8_t VAR_ALL_HOOK_MASK 	= 15 << 4;          // = 11110000

/**Helper class to watch particular clauses.*/
class WatchedElement{
    /**For binary clauses the rightmost bit is 1 -> during propagation watched cls does not have to be loaded
    * for larger clauses the rightmost bit is 0, the literal is the other watched literal in the clause --> used for
    * fast detection of satisfied clauses*/
    uint32_t literal31_bin1;

    /**The clause to be watched.*/
    Clause*  cls;

public:



    /**Create a watcher for the specified clause.
     * @param lit: either the first or the second literal in the clause.
     * @param cl: the clause to be watched
     * @param binary: true if the watched clause is a binary clause*/
    inline WatchedElement(const Literal& lit);
    inline WatchedElement(const Literal& lit,Clause* cl);

    /**Assignment*/
    inline WatchedElement& operator=(const WatchedElement& copy);

    /**Default construction for std::vector*/
    WatchedElement()                         {WatchedElement(INVALID_LIT,NULL);}

    /**Copy construction*/
    WatchedElement(const WatchedElement& copy):
            literal31_bin1(copy.literal31_bin1),cls(copy.cls){}


    /**Returns the watched literal within the watched clause*/
    Literal  getLiteral()                    const   {return Literal(literal31_bin1 >> 1);}

    /**Indicates if the watched clause is binary.*/
    bool     isBinary()                      const   {return literal31_bin1 & 1;}

    Clause*  getClause()                     const   {asrt(!isBinary());return cls;}

    /**Sets the literal to be watched in the clause.*/
    void     setLiteral(const Literal& lit){
           literal31_bin1 = (lit.idx() << 1) | (literal31_bin1 & 1);}


    struct ClauseEqualFunctor{
     	const Clause* comp;
     	ClauseEqualFunctor(Clause* compare):comp(compare){asrt(compare != NULL);};
        bool operator() (const WatchedElement& e){
        	asrt(!e.isBinary() || e.cls == NULL);
        	return e.cls == comp;}
    };

    struct BinaryEqualFunctor{
    	const uint32_t comp;
        BinaryEqualFunctor(Literal compare):comp((compare.idx()<<1)|1){};
        bool operator() (const WatchedElement& e){return e.literal31_bin1 == comp;}
    };
};

typedef vector<WatchedElement> we_vec;
typedef	vector<WatchedElement>::iterator we_iter;
typedef vector<Clause*> ::iterator cls_iter_t;



class AssignInfo{
    /**Basically the level at which a variable was assigned. The rightmost bit is set in case of a binary reason:
     * 1: Variable is assigned with a binary clause as reason (simply the other lit of the clause is stored)
     * 0: Variable is assigned with a normal clause as reason (clause might be NULL for no reason) */
    uint32_t level31_b1;
    union{uint32_t lit; Clause* cls;} reason;

public:

    AssignInfo()                                       :level31_b1(0)              {reason.cls = NULL;}
    AssignInfo(const uint32_t level,Clause* c = NULL)  :level31_b1(level<<1)       {reason.cls = c;}
    AssignInfo(const uint32_t level,const Literal& l)  :level31_b1((level<<1)|1)   {reason.lit = l.idx();}


    void clear()                                 { level31_b1 = 0; reason.cls = NULL;}

    bool isBinReason()           const           { return level31_b1 & 1;}
    bool isClsReason()           const           { return !(level31_b1 & 1);}
    bool isReason(Clause* cl)    const           { return !(level31_b1 & 1) && cl == reason.cls;};
    bool noReason()              const           { return !(level31_b1 & 1) && reason.cls == NULL;};

    Literal      getBinReason()  const           { asrt(isBinReason());return Literal(reason.lit);}
    Clause*      getClsReason()  const           { asrt(isClsReason());return reason.cls;}
    uint32_t     getLevel()      const           { return level31_b1 >> 1;}

    void setBinReason(const uint32_t level,const Literal& l){
           level31_b1  = (level << 1) | 1;
           reason.lit  = l.idx();}

    void setClsReason(const uint32_t level,Clause* c = NULL){
           level31_b1  = level << 1;
           reason.cls  = c;}
};

/**Class to store all information about the formula to be solved. */
class ConstraintDB
{
//general
protected:
    Settings&                       settings;

    Statistics&                     statistics;

    /**Constructor*/
    inline                      ConstraintDB(Settings& setting, Statistics& stats);

       /**Destructor*/
    inline                      ~ConstraintDB();

    typedef std::list<Clause*> uwc_t;
    typedef uwc_t::iterator uwc_iter;
    uwc_t								unwatched_clauses;
    uint32_t							nbr_learnt_uw_clses;




//Variable related Information
private:
    /**An assignment value for each variable and the previous (false) assignment
     * one uint8_t value <abcdefgh> stores 4 different assignments for one variable:
     * ab = hook of a variable (used in walksat)
     * cd = unused
     * ef = previous variable value (assigned during backjumping)
     * gh = value of a variable used during dpll
     * */
    vector< uint8_t      >          assignments;

    /**An assignment level for each variable*/
    vector< AssignInfo   >          assign_info;


protected:
	/**Tells the solver to reserve enough space to hold the given number of variables.*/
	void                        reserveForVariables(uint32_t sz);

	/**Assign a value to the variable without checking the current value.*/
	inline void                 varUncheckedAssign(const Literal& lit,Clause* reason,uint32_t lev);

	inline void					varAssignLitAt0(const Literal& lit){
		const uint32_t   var_idx    = lit.toVar().idx();
		assignments     [var_idx]   = (~VAR_VALUE_MASK & assignments[var_idx]) | lit.pol().toInt();
		assign_info     [var_idx]	.setClsReason(0,NULL);
	}

	inline void                 varUncheckedAssign(const Literal& lit, const Literal& reason, uint32_t lev);

	inline void                 varSetHook(const Literal& lit);

	inline void                 varSetHookFromValue(const Variable& var);

    inline void                 varSetHook(const Variable& var, TrBool val);

    //debug
    inline void					varTakeHookAsFinalValue(const Variable& var);
public:

    /**Simplifier specific*/
    inline void 				increaseFixLiteralsCount(uint cl_sz);

	/**Clear assignment of a variable*/
	inline void                 varClearAssignmentSetPrev(const Variable& v);
	inline void                 varSetPrev(const Variable& v,const TrBool& val);
	inline void                 varClearAssignmentSetPrevFlp(const Variable& v);
	inline void                 varClearAssignmentSimple(const Variable& v);
	inline void        			varClearAssignmentSetPrevAndHook(const Variable& v);

	/**Clear assignment of a variable*/
	//inline void                 varClearHook(const Variable& v);

	/**Return the reason for the assignment of a variable.*/
	inline const AssignInfo&    varAssignInfo(const Variable& v)                    const;

public:
    /**Return the current value of a variable*/
    inline TrBool               varValue(const Variable& v)                         const;

    /**Return the previous value of a variable */
    inline TrBool               varPrevValue(const Variable& v)                     const;

    /**Return the current hook value (walksat) of a variable*/
    inline TrBool               varHook(const Variable& v)                          const;

    /**Return the level of assignment of a variable. For unassigned variables this is INVALID_LEVEL*/
    inline const uint32_t        varAssignLevel(const Variable& v)                   const;



//Literal related Information
protected:
    /**For each literal it holds a vector of clauses/literals that are watched*/
    vector<vector<WatchedElement> > watches;

protected:

	//vector<vector<WalkClInfo*> >  	occurrences;

    /**Return the Watched elements (clauses) of the specified literal*/
    inline vector<WatchedElement>&  litWatchedClauses(const Literal& lit);

    /**Add a clause with size at most 3 to be watched to the specified literal.
     * @param lit: literal to watch the clause
     * @param other_watched: the literal of the other watcher in the clause
     * @param cl: clause to be watched*/
    inline void                     litAddWatchedClause(const Literal& lit, const Literal& other_watched, Clause* cl);

public:
    /**Checks if the specified literal is true under the current assignment.*/
    inline bool                     litIsTrue(const Literal& l)                         const;

    /**Retruns true if the value of the literal is true or unknown under the current assignement*/
    inline bool                     litIsNotFalse(const Literal& l)                     const;

    /**Returns true if the value of the literal is false under the current assignment*/
    inline bool                     litIsFalse(const Literal& l)                        const;

    /**Returns true if the value of the literal is unknown under the current assignment*/
    inline bool                     litIsUnknown(const Literal& l)                      const;

    /**Returns true if the hook value of the variable equals the polarity of this literal*/
    inline bool                     litIsHooked(const Literal& l)                       const;

    /**Print all watched clauses of the specified literal*/
    inline void                     litPrintWatches(const Literal& l)                   const;

    inline const uint32_t            litAssignLevel(const Literal& l)                    const;

    //inline void						unlinkAllClauses();



//Clause related Information
//all arrays store according information to one clause each - indexed by the ID of a clause
protected:
    /**All binary clauses, learnt and original*/
    //vector<Clause*>           	    bin_cls;

    /**Holds a pointer to all clauses that are either original clauses or learnt binaries*/
    vector<Clause*>           		fix_cls;

    /**Holds a pointer to all learnt clauses with size > 3.*/
    vector<Clause*>             	learnt_cls;

private:
    uint32_t                    	nbr_bin_clses;
    uint64_t						nbr_long_learnt_literals;
    uint64_t						nbr_long_fix_literals;

protected:
	cls_act_t   	  				cls_inc;



	/**Delete and remove the specified clause.*/
	void                       		clsUncreate(Clause* cl,bool unlink , bool free_mem );

	inline bool                		clsLocked(Clause& cls);


protected:
    /**Create a clause from the specified literals (at least 3!!).
     * @param lits: literals of the clause to be created
     * @param learnt: indicates if the clause to create is a learnt clause
     * @return the newly created clause*/
    Clause*						clsCreate(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend, bool learnt = false, bool linked = true);

    void						clsCreateBinary(const Literal lit0, const Literal lit1);

    inline bool					containsBinary(const Literal lit0, const Literal lit1);

    inline void    				clsUnlinkWatched(Clause* cl,const bool keep_unlinked_list);
    inline void    				clsIrregularUnlink(Clause* cl,we_vec& from);
    inline void    				clsLinkWatched(Clause* cl);


    /**Increment the activity value of the specified clause by the given amount.*/
    inline void					clsBump(Clause* cl);
    inline void					clsWeakenAll(vector<Clause*>& clauses,const cls_act_t::denominator_t w );

    inline void					clsClearAllStamps();


    inline void					clsShrinkToSize(Clause& cl, uint32_t sz);
    inline void					clsRemoveLiteral(Clause& cl, uint32_t index);

    uint32_t 					keepProblemClauses(uint32_t cut, std::list<Clause*>& out_container);

     // Reduce the set of learnt clauses.
    void						learntsGC(uint32_t new_sz,bool free_mem = true);

    void						removeSat(vector<Clause*>& clauses);


public:
    /**Tell the ConstraintDB instance to reserve enough space to hold the given number of clauses.*/
    inline void                reserveForClauses(uint32_t sz);


//general
public:

    uint32_t numberVariables()          		const   {return assignments.size();}
	uint64_t numberLiteralsInLongLearntClses()	const	{return nbr_long_learnt_literals;}
	uint64_t numberLiteralsInLongFixClses()		const	{return nbr_long_fix_literals;}
	uint64_t numberLiteralsInLongClses()		const	{return numberLiteralsInLongLearntClses()+numberLiteralsInLongFixClses();}
	uint64_t numberLiteralsInShortClses()		const	{return nbr_bin_clses<<1;}
	uint64_t numberLiterals()					const	{return numberLiteralsInLongClses()+numberLiteralsInShortClses();}

	uint32_t numberBinClses()					const	{return nbr_bin_clses;}
	uint32_t numberLearntLongClses()			const	{return learnt_cls.size();}
	uint32_t numberOrigLongClses()				const	{return fix_cls.size();}
	uint32_t numberAllLongClses()				const	{return numberOrigLongClses() + numberLearntLongClses();}
	uint32_t numberAllClauses()					const	{return numberAllLongClses() + numberBinClses();}


    bool 						clsCanBeOmitted(Clause& cl);

    /**Check if the given clause is satisfied under the current assignment.*/
    inline Literal              checkSatByLit(Clause& cl)                                  const;

    inline Literal              checkSatByHook(Clause& cl)                                  const;



    /**Write the current set of clauses in DIMACS format.
     * @param skip_sat_cls: if true only clauses are printed that are not satisfied under the current assignment.*/
    inline  void                writeDimacs(OFStream& out, bool skip_sat_cls=false)    const;

	void						printModel();
	void						printHookModel();

    //Debug
    bool debugVerifyModel();
    bool debugVerifyModelByHooks();

	inline bool	debugAtLeastOneHooked(Clause& cl);
	inline bool	debugAtLeastOneHooked(vector<Literal>& lits,uint start);
    inline bool debugTestNumberUnwatched(uint32_t should_be);
    inline bool debugRangeFilterFlagged(cls_iter_t begin, cls_iter_t end);

};





//Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#define ch_idx(index,vector) asrt(index.ID >= 0 && (uint32_t)index.ID < vector .size() )


//Implementation WatchedElement
WatchedElement::WatchedElement(const Literal& lit):
	literal31_bin1((lit.idx() << 1) + 1),cls(NULL){}

WatchedElement::WatchedElement(const Literal& lit,Clause* cl):
    literal31_bin1((lit.idx() << 1) ),cls(cl){}

WatchedElement& WatchedElement::operator=(const WatchedElement& copy){
    literal31_bin1  = copy.literal31_bin1;
    cls             = copy.cls;
    asrt(!copy.isBinary() || copy.cls == NULL);
    return *this;
}


void		ConstraintDB::increaseFixLiteralsCount(uint cl_sz)		{nbr_long_fix_literals += cl_sz;}

TrBool      ConstraintDB::varValue(const Variable& v)       const   {ch_idx(v,assignments);return TrBool(assignments[v.ID]&VAR_VALUE_MASK);}
TrBool      ConstraintDB::varPrevValue(const Variable& v)   const   {ch_idx(v,assignments);return TrBool((assignments[v.ID]>>2)&VAR_VALUE_MASK);}
TrBool      ConstraintDB::varHook(const Variable& v)        const   {ch_idx(v,assignments);return TrBool(assignments[v.ID]>>6);}
bool        ConstraintDB::litIsTrue(const Literal& l)       const   {return varValue(l.toVar()) == l.pol();}
bool        ConstraintDB::litIsNotFalse(const Literal& l)   const   {return varValue(l.toVar()) != l.flpdPolarity();}
bool        ConstraintDB::litIsFalse(const Literal& l)      const   {return varValue(l.toVar()) == l.flpdPolarity();}
bool        ConstraintDB::litIsUnknown(const Literal& l)    const   {return varValue(l.toVar()).isUnknown();}
bool        ConstraintDB::litIsHooked(const Literal& l)     const   {return varHook (l.toVar()) == l.pol();}

void        ConstraintDB::varClearAssignmentSetPrev(const Variable& v) {
	 assignments[v.ID]   = (assignments[v.ID]                              & VAR_ALL_HOOK_MASK)    //first four bits remain unchanged
					   | (((assignments[v.ID] << 2) | UNKNOWN_TRBOOL_VAL)  & VAR_ALL_VAL_MASK);   //last four bits are changed
}

void        ConstraintDB::varSetPrev(const Variable& v,const TrBool& val) {
	 assignments[v.ID]   = (assignments[v.ID]	& (VAR_ALL_HOOK_MASK|VAR_VALUE_MASK))    //first four bits remain unchanged and last two
					   | (val.toInt() << 2);   //bit 2,3 are changed
	asrt(varPrevValue(v)==val);
}

void        ConstraintDB::varClearAssignmentSetPrevFlp(const Variable& v)  {
	ch_idx(v,assignments);
	asrt(TrBool(assignments[v.ID]&VAR_VALUE_MASK).isKnown());
	assignments[v.ID] = (assignments[v.ID]                              						& VAR_ALL_HOOK_MASK)    //first four bits remain unchanged
					  | ((((assignments[v.ID] ^ TRBOOL_FLIP_MASK) << 2) | UNKNOWN_TRBOOL_VAL) 	& VAR_ALL_VAL_MASK );
}

void        ConstraintDB::varClearAssignmentSimple(const Variable& v)     {
	ch_idx(v,assignments);
    assignments[v.ID] = (assignments[v.ID] & VAR_VALUE_MASK_INV) | UNKNOWN_TRBOOL_VAL;
}


void        ConstraintDB::varClearAssignmentSetPrevAndHook(const Variable& v)     { ch_idx(v,assignments);
    assignments[v.ID]   = (assignments[v.ID]   << 6                         & VAR_ALL_HOOK_MASK)    //first two bits are set as hook
                        | (((assignments[v.ID] << 2) | UNKNOWN_TRBOOL_VAL)  & VAR_ALL_VAL_MASK);   //last four bits are changed
}

/*
void        ConstraintDB::varSetHookFromValue(const Variable& v){
	assignments[v.ID]   = (((assignments[v.ID] >>2)|(assignments[v.ID]<<6))	& VAR_ALL_HOOK_MASK)    //first four bits are changed
	                    | (assignments[v.ID]                            	& VAR_ALL_VAL_MASK);   //last four bits remain unchanged
}*/
/*void        ConstraintDB::varClearHook(const Variable& v)           { ch_idx(v,assignments);
    assignments[v.ID]   = (((assignments[v.ID] >>2)|(UNKNOWN_TRBOOL_VAL<<6))& VAR_ALL_HOOK_MASK)    //first four bits are changed
                        | (assignments[v.ID]                                & VAR_ALL_VAL_MASK);   //last four bits remain unchanged
}*/

const AssignInfo& ConstraintDB::varAssignInfo(const Variable& v)const   {
    ch_idx(v,assign_info); return assign_info[v.ID];}



const uint32_t  ConstraintDB::varAssignLevel(const Variable& v) const
            {ch_idx(v,assign_info); asrt(varValue(v).isKnown());return assign_info[v.ID].getLevel();}
const uint32_t  ConstraintDB::litAssignLevel(const Literal& l)  const
            {ch_idx(l.toVar(),assign_info); asrt(!litIsUnknown(l)); return assign_info[l.toVar().ID].getLevel();}




void        ConstraintDB::clsBump(Clause* cl){

    if((cl->getActivity() += cls_inc) > settings.cls_max_act ) {
        statistics.nbr_expensive_cls_bmp += 1;
        const cls_act_t::denominator_t w = settings.cls_min_inc.getLogOfDenominator();
        cls_inc.weaken(w);
        for(int i = learnt_cls.size(); i--;) learnt_cls[i]->getActivity().weaken(w);
        for(int i = fix_cls.size();i--;)fix_cls[i]->getActivity().weaken(w);
    }
}

/*
Clause*		ConstraintDB::getNextLearnt(uint32_t& idx) const {
	while(idx< learnt_cls.size()){
		Clause* ret = learnt_cls[idx++];
		if(!ret->isRubbish()) return ret;
	}
	return NULL;
}

Clause*		ConstraintDB::getNextOrig(uint32_t& idx) const {
	while(idx< fix_cls.size()){
		Clause* ret = fix_cls[idx++];
		if(!ret->isRubbish()) return ret;
	}
	return NULL;
}*/



template<class C>
void free_all(vector<C>& vec){
    int32_t i(vec.size());
    while(i--){
       asrt(vec[i] != NULL);
       free(vec[i]);
       vec[i] =NULL;
    }
}

ConstraintDB::~ConstraintDB(){
    free_all(learnt_cls);
    free_all(fix_cls);
};

ConstraintDB::ConstraintDB(Settings& sets, Statistics& stats):
	settings(sets),
	statistics(stats),
	nbr_learnt_uw_clses(0),
	nbr_bin_clses(0),
	nbr_long_learnt_literals(0),nbr_long_fix_literals(0),
	cls_inc(sets.cls_min_inc){ }

void        ConstraintDB::varSetHook(const Variable& var, TrBool val){
	assignments     [var.idx()]   = (val.toInt() << 6) | (~VAR_HOOK_MASK & assignments[var.idx()])  ;
	asrt(varHook(var) == val);
}

void        ConstraintDB::varSetHook(const Literal& lit){
	varSetHook(lit.toVar(),lit.pol());}

void        ConstraintDB::varUncheckedAssign(const Literal& lit, Clause* reason, uint32_t lev){
    asrt(litIsUnknown(lit));

    const uint32_t   var_idx    = lit.toVar().idx();
    assignments     [var_idx]   = (~VAR_VALUE_MASK & assignments[var_idx]) | lit.pol().toInt();
    assign_info     [var_idx].setClsReason(lev,reason);
    asrt(assign_info[var_idx].getLevel() == lev);
    asrt(assign_info[var_idx].getClsReason() == reason);
}

void        ConstraintDB::varUncheckedAssign(const Literal& lit, const Literal& reason, uint32_t lev){
    asrt(litIsUnknown(lit));

    const uint32_t   var_idx    = lit.toVar().idx();
    assignments     [var_idx]   = (~VAR_VALUE_MASK & assignments[var_idx]) | lit.pol().toInt();
    assign_info     [var_idx].setBinReason(lev,reason);
    asrt(assign_info[var_idx].getLevel() == lev);
    asrt(assign_info[var_idx].getBinReason() == reason);
}




void    ConstraintDB::reserveForClauses(uint32_t need_sz){
    if(need_sz > fix_cls.size()){fix_cls.reserve(need_sz);}
}


void    ConstraintDB::litAddWatchedClause(const Literal& lit, const Literal& other_watched,Clause* cl){
    ch_idx(lit,watches); asrt(cl != NULL && cl->size() > 2);
    asrt(((*cl)[0] == lit && (*cl)[1] == other_watched) || ((*cl)[1] == lit && (*cl)[0] == other_watched));
    watches[lit.idx()].push_back(WatchedElement(other_watched,cl));
}

void	ConstraintDB::clsClearAllStamps(){
	for(uint32_t i = fix_cls.size();i--;){fix_cls[i]->clearStamp();}
	for(uint32_t i = learnt_cls.size();i--;){learnt_cls[i]->clearStamp();}
}




bool ConstraintDB::containsBinary(const Literal lit0, const Literal lit1){
	bool contained = find(litWatchedClauses(lit0),WatchedElement::BinaryEqualFunctor(lit1));
	asrt(contained == find(litWatchedClauses(lit1),WatchedElement::BinaryEqualFunctor(lit0)));
	return contained;
}



void    ConstraintDB::clsLinkWatched(Clause* cl){
    const Literal lit0 = (*cl)[0];
    const Literal lit1 = (*cl)[1];

    asrt(!cl->isWatched());
    asrt(!find(litWatchedClauses(lit0),WatchedElement::ClauseEqualFunctor(cl)));
    asrt(!find(litWatchedClauses(lit1),WatchedElement::ClauseEqualFunctor(cl)));

    litAddWatchedClause(lit0, lit1,cl);
    litAddWatchedClause(lit1, lit0,cl);

    cl->setWatched();
}


void    ConstraintDB::clsUnlinkWatched(Clause* cl,const bool keep_unlinked_list){
	asrt(cl != NULL);
    WatchedElement::ClauseEqualFunctor cl_equal(cl);

    asrt(cl->isWatched());
    asrt(find(litWatchedClauses((*cl)[0]),cl_equal));
    stable_remove_one(litWatchedClauses((*cl)[0]),cl_equal);

    asrt(find(litWatchedClauses((*cl)[1]),cl_equal));
    stable_remove_one(litWatchedClauses((*cl)[1]),cl_equal);
    cl->unsetWatched();

    if(keep_unlinked_list){
    	unwatched_clauses.push_front(cl);
    	if(cl->isLearnt())	++nbr_learnt_uw_clses;
    }
}

void    ConstraintDB::clsIrregularUnlink(Clause* cl,we_vec& from){
	asrt(cl != NULL);
    asrt(cl->isWatched());

	asrt(find(from,WatchedElement::ClauseEqualFunctor(cl)));
	stable_remove_one(from,WatchedElement::ClauseEqualFunctor(cl));

	cl->unsetWatched();
	unwatched_clauses.push_front(cl);
	if(cl->isLearnt())	++nbr_learnt_uw_clses;
}





void		ConstraintDB::clsShrinkToSize(Clause& cl, uint32_t sz){
	asrt(sz < cl.size());
	if(cl.isLearnt())	nbr_long_learnt_literals 	-= (cl.size() - sz);
	else 				nbr_long_fix_literals		-= (cl.size() - sz);
	cl.shrinkToSize(sz);
}

void	ConstraintDB::clsRemoveLiteral(Clause& cl, uint32_t index){
	asrt(index < cl.size());
	statistics.nbr_removed_cls_lits += 1;
	cl.removeAt(index);
	if(cl.isLearnt())	--nbr_long_learnt_literals;
	else				--nbr_long_fix_literals;
}



Literal ConstraintDB::checkSatByLit(Clause& cl) const{
    for(uint32_t i = 0; i != cl.size(); ++i) {
        if(litIsTrue(cl[i])) return cl[i];
    }
    return INVALID_LIT;
}

Literal ConstraintDB::checkSatByHook(Clause& cl) const{
    for(uint32_t i = 0; i != cl.size(); ++i) {
    	asrt(varHook(cl[i].toVar()).isKnown());
        if(litIsHooked(cl[i])) return cl[i];
    }
    return INVALID_LIT;
}



void ConstraintDB::writeDimacs(OFStream& out, bool skip_sat_cls) const{
    //print_cout("p cnf %i %i",numberOfVariables(),numberOfClauses());
    for(uint32_t i = 0; i != fix_cls.size(); ++i){
        Clause* cl = fix_cls[i];
        if(skip_sat_cls && checkSatByLit(*cl) != INVALID_LIT) continue;

        cl->print(out);
    }
}


vector<WatchedElement>& ConstraintDB::litWatchedClauses(const Literal& lit)
    {ch_idx(lit,watches); return watches[lit.ID];}

bool ConstraintDB::clsLocked(Clause& cls){
	return assign_info[cls[0].toVar().ID].isReason(&cls) ? litIsTrue(cls[0])
		   : (assign_info[cls[1].toVar().ID].isReason(&cls) && litIsTrue(cls[1]));
}


#undef ch_idx

const char*  constraintdb_cpp_version();
#define constraintdb_hpp_version "ConstraintDB.hpp:   $Revision: 1.2 $"



bool	ConstraintDB::debugAtLeastOneHooked(Clause& cl){
	DEBUG_FUNC
	for(uint i = 0; i != cl.size();++i){
		if(litIsHooked(cl[i])) return true;
	}
	cl.print(*settings.console_out);
	return false;
}

bool	ConstraintDB::debugAtLeastOneHooked(vector<Literal>& lits,uint start){
	DEBUG_FUNC

	for(; start < lits.size();++start){
		if(litIsHooked(lits[start])) return true;
	}
	return false;
}

/*
bool ConstraintDB::debugRangeFilterFlagged(cls_iter_t begin, cls_iter_t end){
	while(begin != end){
		if(!(*begin++)->filterFlagged()) return false;
	}
	return true;
}*/



bool ConstraintDB::debugTestNumberUnwatched(uint32_t should_be){
	uint32_t sum = 0;
	for(uint i = 0; i!= fix_cls.size();++i){if(!fix_cls[i]->isRubbish() && !fix_cls[i]->isWatched()) ++sum;}
	for(uint i = 0; i!= learnt_cls.size();++i){if(!learnt_cls[i]->isRubbish() &&!learnt_cls[i]->isWatched()) ++sum;}

	return sum == should_be ;
}



#endif /*CONSTRAINTDB_HPP_*/
