//============================================================================
// Name        : SApperloT.cpp
// Author      : Stephan Kottler
//============================================================================

#include "Statistics.hpp"
#include "DMRPSolver.hpp"
#include "BasicSolver.hpp"
#include <signal.h>

#include "Types.hpp"
#include "IFStream.hpp"




/**
 * Some checks
 */

bool check_defines(){

	std::cout << "Asserts will be evaluated and wordsize is " << sizeof(void*) << std::endl;
    return
    //sizeof(void*)           ==      32/8                &&
    sizeof(1)				==		32/8				&&
    sizeof(int32_t)         ==      32/8                &&
    sizeof(int64_t)         ==      64/8                &&
    sizeof(uint32_t)        ==      32/8                &&
    sizeof(uint64_t)        ==      64/8                &&
    sizeof(double)          ==      64/8                &&
    sizeof(float)           ==      32/8                &&
    sizeof(Literal)         ==      sizeof(uint32_t)    &&
    sizeof(Variable)        ==      sizeof(uint32_t)    &&
    sizeof(float)           ==      sizeof(uint32_t)    &&
    ((int)false)            ==      0                   &&
    ((int)true)             ==      1;
};



Settings    settings;
Statistics  statistics(settings);
DMRPSolver solver(settings,statistics);
//BasicSolver solver(settings,statistics);


static void SIGQUIT_handler(int signum){

   /* OFStream out(stderr);
    out.writeNewLine();
    out.setBackgroundPattern(   "|                                                                                            |\n");
    out.writeString(            "|--------------------------------------------------------------------------------------------");
    out.writeCenteredToBG("!!! INFO !!!");
    statistics.printFinalStatistics(out,&solver);
    out.writeCenteredToBG("!!! INFO !!!");
    out.writeString(            "|--------------------------------------------------------------------------------------------");
*/
	exit(EXIT_UNKNOWN);
};

static void SIGINT_handler(int signum) {
    /*OFStream out(stderr);
    out.writeNewLine();
    out.setBackgroundPattern(   "|                                                                                            |\n");
    out.writeString(            "|--------------------------------------------------------------------------------------------");
    out.writeCenteredToBG("!!! KILLED !!!");
    statistics.printFinalStatistics(out,&solver);
    out.writeCenteredToBG("!!! KILLED !!!");
    out.writeString(            "|--------------------------------------------------------------------------------------------");
    */
    exit(EXIT_UNKNOWN);
};


bool parseDIMACS(IFStream& in, BasicSolver& solver) {
    vector<Literal> parsed_lits;
    int32_t  parsed_lit;
    Literal lit;

    uint32_t numberOfLiterals = 0;

    while(in.hasNext()){
		if(in.currEquals('c')) {in.skipLine();}
		else if(in.currEquals('p')){
			if (in == "p cnf"){
				int32_t variables = in.getNextInt();

				int32_t clauses = in.getNextInt();

				statistics.nbr_clauses_original = clauses;

				if(clauses >= 3000000)
					settings.simp = false;

				solver.reserveForVariables(variables);
				solver.reserveForClausesSimplifier(clauses);
				in.skipLine();
			}else{ fprintf(stderr,"c PARSE ERROR in file %s \n",settings.cnf_file_name), exit(EXIT_FAILURE); }
		}else{
			while((parsed_lit = in.getNextInt()) != 0){
				lit = Literal::DIMACS2Literal(parsed_lit);
//                    solver.maybeNewVariable(lit.toVar());
				/*if(parsed_lit > 0)
					lit = Variable(parsed_lit).toPosLit();
				else
					lit = Variable(parsed_lit*(-1)).toNegLit();*/

				solver.reserveForVariables(lit.toVar().idx()+1);
				//solver.maybeNewVariable(lit.toVar());
				parsed_lits.push_back(lit);


			}
			if(settings.simp)
			{
				if(!solver.addClauseForSimplifier(parsed_lits)){
					fprintf(stderr,"c Formula contains an empty clause!! - is unsatisfiable!");
					solver.setInvalid(); return false;
				}
			}else
			{
				if(!solver.addClause(parsed_lits))
				{
					fprintf(stderr,"c Formula contains an empty clause!! - is unsatisfiable!");
					solver.setInvalid(); return false;
				}
			}

			numberOfLiterals+= parsed_lits.size();
			parsed_lits.clear();
		}

	}
    statistics.nbr_literals_original = numberOfLiterals;
    return true;
}



int main(int argc, char** argv){
	sanity((print_out("Sanity checks will be evaluated\n"),true));

    asrt(check_defines());

    if(!settings.read(argc,argv)){
        settings.printUsage();
        return EXIT_FAILURE;
    }

    //try to open file
    IFStream in(settings.cnf_file_name);
    //FILE * in = fopen (settings.cnf_file_name, "r");
    if(in.isBad()){
    	fprintf(stderr,"c Could not open CNF file %s ", settings.cnf_file_name);
        return EXIT_FAILURE;
    }

    //reconstruct it here, because the real settings of the simplification are known here properly
    DMRPSolver solver(settings,statistics);

    signal(SIGINT,SIGINT_handler);
    signal(SIGQUIT,SIGQUIT_handler);



    if(settings.console_out != NULL){
    	settings.console_out->writeNewLine();
    	settings.console_out->writeString(
    	    "|############################################################################################|\n");
    	settings.console_out->setBackgroundPattern(
    	    "|                                                                                            |\n");
    }

    parseDIMACS(in,solver);
    in.close();

    statistics.parse_timestmp = statistics.cpuUserTime() ;


    char* result_txt = (char*)"UNKNOWN";

    TrBool result 	= solver.isSAT( settings.timeout);
    int	ret_val		= EXIT_UNKNOWN;

    if(result.isKnown()){
    	printf("c Solved by %s \n", solver.whoSolvedInstane());

    	if(result.isTrue()){
			result_txt 	= (char*)"SAT";
			ret_val 	= EXIT_SAT;
			printf("\ns SATISFIABLE\n");
			if(settings.printmodel)
				solver.printModel();
		}else{
			result_txt =(char*)"UNSAT";
			ret_val 	= EXIT_UNSAT;
			printf("\ns UNSATISFIABLE\n");
		}

    }else{
    	printf("\ns UNKNOWN\n");
    }



    if(settings.csv_file_name != NULL){
        OFStream out(settings.csv_file_name,false);
        statistics.printCSV(out,&solver,result_txt);
    }


    if(settings.console_out != NULL){
    	settings.console_out->writeString(
    	    "|Instance %s is %s", bname(settings.cnf_file_name),result_txt );
    	statistics.printFinalStatistics(*settings.console_out,&solver);
    	settings.console_out->writeString(
    	    "|############################################################################################|\n");

    }



#ifdef NDEBUG
	exit(ret_val);
#endif

    return ret_val;
}




const char*  sapperlot_cpp_version(){return "SApperloT.cpp:      $Revision: 1.2 $";}


