#ifndef SATTYPES_HPP_
#define SATTYPES_HPP_

#include <vector>
#include "Assertion.hpp"
#include <stdio.h>
#include <stdint.h>
#include "LazyFraction.hpp"
#include "OFStream.hpp"
#include "Settings.hpp"


using std::vector;


class Literal;
class Clause;

const uint32_t zero32 	= 0;
const uint64_t zero64 	= 0;
const uint64_t max64	= ~zero64;

/**Class to represent boolean values and a third state 'Unknown'*/
class TrBool{
    /**The internal value*/
    uint8_t val;

//public members
public:
    /**Constructs an object with unknown value*/
    inline              TrBool();

    /**Constructs an object with the specified value.
     * @param i: a value from {0,1,2}*/
    inline explicit     TrBool(uint8_t i);

    /**Copy construction*/
    inline              TrBool(const TrBool& b);

    inline uint8_t      toInt()                     const;
    inline bool         isTrue()                    const;
    inline bool         isFalse()                   const;
    inline bool         isUnknown()                 const;
    inline bool         isKnown()                   const;
    inline bool         operator==(const TrBool& o) const;
    inline bool         operator!=(const TrBool& o) const;
};


/**Class to represent variables. All additional information about a variable are stored outside: in an instance of
 * ConstraintDB*/
class Variable{

	uint32_t ID;

public:

    /**Construct a variable from the given id*/
    inline explicit     Variable(const uint32_t i);

    inline uint32_t     idx()							const;
    inline Literal      toPosLit()						const;
    inline Literal      toNegLit()						const;
    inline uint         toDIMACS()						const;
    inline bool         operator<(const Variable& v)	const;
    inline bool         operator!=(const Variable& v)	const;
    inline bool         operator==(const Variable& v)	const;

    friend class Literal;
    friend class ConstraintDB;
};


/**Class to represent literals. All aditional information are stored elsewhere: in an instance of class ConstraintDB*/
class Literal{

	uint32_t ID;

public:
    /**Return a literal object from the given DIMACS number*/
    inline static Literal DIMACS2Literal(int32_t d);

    /**Construct an invalid literal*/
    inline                  Literal();

    inline explicit         Literal(const uint32_t i);

    inline uint32_t         idx()			const;
    inline Literal          flipped()		const;
    inline Variable         toVar()			const;
    inline bool             isPos()			const;
    inline bool             isNeg()			const;
    inline TrBool           pol()			const;
    inline TrBool           flpdPolarity()       const;
    inline int              toDIMACS()              const;
    inline bool             operator<(const Literal& l) const;
    inline bool             operator!=(const Literal& l) const;
    inline bool             operator==(const Literal& l)const;

    //for simplifier
    inline uint32_t 		hash()			const;

    friend class Variable;
    friend class ConstraintDB;
};


/**Class to represent clauses. Contrary to the other types each clause is represented by exactly one instance of this class Clause.
 * Hence, we deal with pointers of this class. Note that all instances of this class can have different sizes although sizeof(Clause)
 * will always return the same value.*/
class Clause{
public:
	static const uint32_t MAX_SIZE 	= (1 << 30) - 2;
	static const uint32_t MAX_STAMP = (1 << 20) - 1;

	/**id used during simplification steps */
private:
	union{uint32_t myID;  uint32_t activity;} quality;


    /**Three different information are hold here: The size of the clause is stored in the 29 leftmost bits.
     * The two rightmos bits are reserved to mark this clause and the remaining bit is set iff this clause is learnt.*/
    uint32_t size29_watched_flag_learnt;

    union{uint32_t nextk12_stamp20; uint32_t signature;} state;

    /**The literals start here. The fields are filled from outside the class.*/
    Literal literals[0];

public:
    /**Construct a part of the clause*/
	inline Clause(const uint32_t& sz, bool learnt, uint32_t idx);
public:
	cls_act_t& getActivity(){return (cls_act_t&)quality.activity;}

//private members
private:
    /**Construct a part of the clause*/
    inline Clause(const uint32_t& sz, bool learnt);

    /**Copying of Clause instances is not allowed*/
    Clause(const Clause& cl){asrt(false);} // no copying
    Clause& operator=(const Clause& c){asrt(false);return (Clause&)c;} // no assignment

public:
	struct ActivityDescendingFunctor{
	  	bool operator() (Clause* a, Clause* b){	return a->getActivity()  > b->getActivity();	}
	};


    inline const uint32_t   size()					const;
    inline cls_act_t        getActivity()			const;
    inline const bool       isLearnt()				const;
    //inline const bool       filterFlagged()			const;
    //inline void             setFilterFlag();
    //inline void             unsetFilterFlag();
    inline const bool       isWatched()				const;
    inline void             setWatched();
    inline void             unsetWatched();


    /**Access the literals of a clause by an index*/
    inline Literal&         operator [] (uint32_t i);

    inline void             print(OFStream& out);
    inline void				removeAt(uint32_t idx);
    inline void				shrinkToSize(uint32_t new_sz);
    inline uint32_t 		getNextK(const uint32_t& stamp);
    inline void 			setNextK(const uint32_t k, const uint32_t& stamp);
    inline void				clearStamp();
    inline void				clearStampAndFlag();
    inline bool				isRubbish() const;

    inline uint32_t getSignature()					const;
    inline void computeSignature()					;

	inline void flipLiteral(uint index)				;
	inline void eliminateLiteral(Literal toElim)	;

	inline void del()								;
	inline bool null()								const;

	inline uint32_t getID()							const;
	inline void setID(uint32_t id)					;

	inline vector<Literal> getAllLits()				;

private:
    inline void				setRubbish();

    friend class ConstraintDB;
};


//Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
const Literal   INVALID_LIT             (~zero32);
const Variable  INVALID_VAR             (~zero32);

const uint8_t   FALSE_TRBOOL_VAL      = 0;
const uint8_t   TRUE_TRBOOL_VAL       = 1;
const uint8_t   TRBOOL_FLIP_MASK      = 1;
const uint8_t   UNKNOWN_TRBOOL_VAL    = 2;
const TrBool    TRUE_TrBool(TRUE_TRBOOL_VAL);
const TrBool    FALSE_TrBool(FALSE_TRBOOL_VAL);
const TrBool    UNKNOWN_TrBool(UNKNOWN_TRBOOL_VAL);



//TBool Implementation
            TrBool::TrBool()                            :val(UNKNOWN_TRBOOL_VAL){ }
            TrBool::TrBool(const TrBool& b)             :val(b.val)             { }
            TrBool::TrBool(uint8_t i)                   :val(i)     {
                            asrt(i == TRUE_TRBOOL_VAL || i == FALSE_TRBOOL_VAL || i == UNKNOWN_TRBOOL_VAL);}

uint8_t     TrBool::toInt()                 const       {return val;}
bool        TrBool::isTrue()                const       {return val == TRUE_TRBOOL_VAL;}
bool        TrBool::isFalse()               const       {return val == FALSE_TRBOOL_VAL;}
bool        TrBool::isUnknown()             const       {return val == UNKNOWN_TRBOOL_VAL;}
bool        TrBool::isKnown()               const       {return val != UNKNOWN_TRBOOL_VAL;}
bool        TrBool::operator==(const TrBool& o) const   {return val == o.val;}
bool        TrBool::operator!=(const TrBool& o) const   {return val != o.val;}



//Variable Implementation
            Variable::Variable(const uint32_t i) :ID(i)          { }
Literal     Variable::toPosLit()                    const   {return Literal((ID << 1) + 1);}
Literal     Variable::toNegLit()                    const   {return Literal(ID << 1);}
uint32_t    Variable::idx()                         const   {return ID;}
uint32_t    Variable::toDIMACS()                    const   {return ID+1;}
bool        Variable::operator<(const Variable& v)  const   {return ID < v.ID;}
bool        Variable::operator!=(const Variable& v) const   {return ID != v.ID;}
bool        Variable::operator==(const Variable& v) const   {return ID == v.ID;}

//Literal Implementation
            Literal::Literal()              		:ID(~zero32){ }
            Literal::Literal(const uint32_t i)   	:ID(i)      { }
uint32_t    Literal::idx()                          const   {return ID;}
Literal     Literal::flipped()                      const   {return Literal(ID ^ 1);}
Variable    Literal::toVar()                        const   {return Variable(ID >> 1);}
bool        Literal::isPos()                        const   {return ID & 1;}
bool        Literal::isNeg()                        const   {return !(ID & 1);}
int         Literal::toDIMACS()                     const   {return ID&1 ? ((ID>>1)+1) : -((ID>>1)+1);}
TrBool      Literal::pol()                          const   {return TrBool(ID & 1);}
TrBool      Literal::flpdPolarity()                 const   {return TrBool(!(ID&1));}
bool        Literal::operator==(const Literal& l)   const   {return ID == l.ID; }
bool        Literal::operator<(const Literal& l)    const   {return ID < l.ID;}
bool        Literal::operator!=(const Literal& l)   const   {return ID != l.ID;}

uint32_t	Literal::hash()							const	{return ID % 32;}

Literal     Literal::DIMACS2Literal(int32_t d){
    asrt(d != 0);
    //return Literal(d < 0 ? (-d-1)*2 : d*2-1);
    uint32_t m(d>>31);
    return Literal(((d^m)<<1)-(m+1));
}


const uint32_t  Clause::size()                      const   {asrt(!isRubbish());return size29_watched_flag_learnt >> 3;}
const bool      Clause::isLearnt()                  const   {asrt(!isRubbish());return size29_watched_flag_learnt & 1;}
//const bool      Clause::filterFlagged()             const   {asrt(!isRubbish());return size29_watched_flag_learnt & 2;}
const bool      Clause::isWatched()             	const   {asrt(!isRubbish());return size29_watched_flag_learnt & 4;}

//void            Clause::setFilterFlag()                     {asrt(!isRubbish());size29_watched_flag_learnt |= 2;};
void            Clause::setWatched()                     	{asrt(!isRubbish());size29_watched_flag_learnt |= 4;};

//void            Clause::unsetFilterFlag()                   {asrt(!isRubbish());size29_watched_flag_learnt &= ~2;};
void            Clause::unsetWatched()                     	{asrt(!isRubbish());size29_watched_flag_learnt &= ~4;};

Literal&        Clause::operator [] (uint32_t i)            {asrt(!isRubbish());asrt(i<size());return literals[i];}
void			Clause::setRubbish()						{size29_watched_flag_learnt = 0;getActivity()=cls_act_t(0);}
bool			Clause::isRubbish()					const 	{return size29_watched_flag_learnt == 0;}



void			Clause::removeAt(uint32_t idx){
	asrt(!isRubbish());
	asrt((idx > 1 || !isWatched()) && idx < size());
	literals[idx] = literals[size() - 1];
	size29_watched_flag_learnt -= 8;
	clearStamp();
}

void			Clause::shrinkToSize(uint32_t new_sz){
	asrt(!isRubbish());
	asrt(!isWatched());
	asrt(new_sz > 2 && new_sz < size());

	size29_watched_flag_learnt = (new_sz<< 3) | (size29_watched_flag_learnt & 7) ;
	clearStamp();
	asrt(size() == new_sz);
}

Clause::Clause(const uint32_t& sz, bool learnt):
    size29_watched_flag_learnt( (sz << 3) | ((uint32_t)learnt))
    {
	getActivity() 	= cls_act_t(0);
	state			.nextk12_stamp20 = 0;}

Clause::Clause(const uint32_t& sz, bool learnt, uint32_t idx):
    size29_watched_flag_learnt( (sz << 3) | ((uint32_t)learnt)){
	quality.myID 	= idx;
	state.signature = 0;
}

void Clause::print(OFStream& out){
	asrt(!isRubbish());
	out.deactivateBackgroundPattern();
    for(uint32_t l = 0; l != size(); ++l){
        out.writeString("%i ",literals[l].toDIMACS() );
    }
    out.writeNewLine();
    out.reactivateBackgroundPattern();
}

uint32_t Clause::getNextK(const uint32_t& stamp){
	asrt(!isRubbish());
	asrt(stamp <= MAX_STAMP);
	const bool valid = (state.nextk12_stamp20 & MAX_STAMP) == stamp;
	return (valid * (state.nextk12_stamp20 >> 20))+ 2;
}

void 	Clause::setNextK(const uint32_t k, const uint32_t& stamp){
	asrt(!isRubbish());
	asrt(stamp <= MAX_STAMP);
	asrt(k >= 2);
	state.nextk12_stamp20 = ((k-2) << 20) | stamp;
}


void	Clause::clearStamp(){state.nextk12_stamp20 = 0;}

void	Clause::clearStampAndFlag(){
	asrt(!isRubbish());
	state.nextk12_stamp20 = 0;
	size29_watched_flag_learnt &= ~2;
}

uint32_t	Clause::getID()						const	{return quality.myID;}
void		Clause::setID(uint32_t id)					{asrt(quality.myID == 0); quality.myID = id;}
uint32_t	Clause::getSignature()				const	{return state.signature;}
void 		Clause::computeSignature()					{
														state.signature = 0;
														for(uint i=0;i<size();i++)
															state.signature |= 1 << literals[i].hash();
														}


void 		Clause::flipLiteral(uint index)				{asrt(index<size());

														//check whether another (flipped) literal in the clause has the same hash value -> than: do not change the signature
														bool anotherTrue = false;
														bool anotherFlipped = false;

														for(uint i=0;i<size();i++)
														{
															if(i != index)
															{
																if(literals[i].hash() == literals[index].hash())
																	anotherTrue = true;

																if(literals[i].hash() == literals[index].flipped().hash())
																	anotherFlipped = true;
															}
														}
														//mask for the change of the signature:
														uint32_t mask = 0;
														if(!anotherTrue)
															mask = 1 << literals[index].hash();

														literals[index] = literals[index].flipped();
														if(!anotherFlipped)
															mask |= 1 << literals[index].hash();

														state.signature ^= mask;
//														computeSignature();
														}
void		Clause::eliminateLiteral(Literal toElim)	{asrt(state.signature & (1 << (toElim.hash())));

														//check whether another literal in the clause has the same hash value -> than: do not change the signature
														bool another = false;

														//eliminate the Lit from the vector<Lit>
														uint i=0;
														while(!(literals[i] == toElim))
														{
															if(literals[i].hash() == toElim.hash())
																another = true;

															i++;

														}

														//possibility1: shift the rest one index to the front
														while(i < size()-1)
														{
															if(literals[i+1].hash() == toElim.hash())
																another = true;

															literals[i] = literals[i+1];
															i++;
														}

														removeAt(size()-1);
//														size29_watched_flag_learnt -= 8;

														//change the signature
														if(!isRubbish() && size() > 0)
//															state.signature ^= 1 << (toElim.hash());
															computeSignature();
														}

void		Clause::del()								{quality.myID = 0;}
bool		Clause::null()						const	{return quality.myID == 0;}


vector<Literal> Clause::getAllLits()					{
														vector<Literal> allLits;
														allLits.reserve(size());
														for(uint i=0;i<size();i++)
															allLits.push_back(literals[i]);

														return allLits;
														}



#define types_hpp_version "Types.hpp:          $Revision: 1.4 $"


#endif /*SATTYPES_HPP_*/
